//
//  GameController.m
//  GamesWithTeams
//
//  Created by Traun Leyden on 3/16/13.
//  Copyright (c) 2013 TL. All rights reserved.
//

#import "GameController.h"
#import "BoardUIView.h"
#import "Game.h"
#import "Player.h"
#import "ChessGame.h"
#import "GameWithTeams.h"
#import "AppDelegate.h"
#import "GameStore.h"

@interface GameController ()

@end

@implementation GameController {
    GameWithTeams* _gameWithTeams;
    
    // hack to solve problem: when we get the notification from the game that a turn was
    // made, how do we know if it was the current player using the app or a remote player.
    // when its the current player, we need to update the field which track the last move
    // by the white or black team.  when its a remote player, the field has already been
    // updated and we don't want to touch it.
    BOOL lastMoveMadeByRemotePlayer;
    
    IBOutlet UILabel* _nextMovePlayerLabel;
    IBOutlet UILabel* _nextMoveTeamLabel;

}

@synthesize boardUIView = _boardUIView;

- (id)initWithNibName:(NSString *)nibNameOrNil
               bundle:(NSBundle *)nibBundleOrNil
                 game:(GameWithTeams *)gameWithTeams
{
    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
    if (self) {
        _gameWithTeams = gameWithTeams;
        [_gameWithTeams addObserver: self forKeyPath: @"turns" options: 0 context: NULL];
        
        // register for events emitted whenever a turn is completed on the GeekGameBoard
        // can happen when local user moves or remote move applied to board
        [[NSNotificationCenter defaultCenter] addObserver: self
                                                 selector: @selector(turnCompleted:)
                                                     name: kTurnCompleteNotification
                                                   object: nil];
        
    }
    return self;
}

- (void)viewDidLoad
{
    [super viewDidLoad];

    [self startGameNamed: @"ChessGame"];
    
    [self updateBoardInteractionStatus];
    
    [self updateNextMoveLabels];
    
    self.navigationItem.title = [[GameStore sharedInstance] username];


}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    [_gameWithTeams removeObserver: self forKeyPath: @"turns"];
    
    [self.boardUIView.game removeObserver: self forKeyPath: @"currentPlayer" ];
    
    [self.boardUIView.game removeObserver: self forKeyPath: @"winner" ];
    
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
    // Dispose of any resources that can be recreated.
}

- (void) startGameNamed: (NSString*)gameClassName
{
    Game *game = self.boardUIView.game;
    
    [game removeObserver: self  forKeyPath: @"currentPlayer"];
    [game removeObserver: self forKeyPath: @"winner"];
    
    if( gameClassName == nil )
        gameClassName = [[game class] description];
    
    [[self boardUIView] startGameNamed: gameClassName];
    
    game = self.boardUIView.game;
    
    [self updateGameBoardStateFromDB];
        
    [game addObserver: self
           forKeyPath: @"currentPlayer"
              options: NSKeyValueObservingOptionInitial
              context: NULL];

    [game addObserver: self
           forKeyPath: @"winner"
              options: 0
              context: NULL]; 
}

- (void) updateGameBoardStateFromDB
{
    
    Game *game = self.boardUIView.game;
    
    // TODO: verify expected player is making the turn
    // TODO: would be nice to have [applyMoveString animation:NO]

    for (NSDictionary *turnDictionary in [_gameWithTeams turnsNonNull]) {
        NSString *lastMove = [turnDictionary objectForKey:@"move"];
        lastMoveMadeByRemotePlayer = YES;  // prevent saving move to DB in game callback
        BOOL moveApplied = [game applyMoveString:lastMove];
        if ( moveApplied == NO )
            NSLog(@"warning: move %@ was not applied to board", lastMove);
        else
            NSLog(@"applied move: %@", lastMove);
            
    }

    
}


- (void) saveMoveToDB: (Turn*)turn
{
    int geekGameBoardPlayerIndex = [[turn player] index];
    TeamId teamId = geekGameBoardPlayerIndex;  // these enums happen to match up in values
    [_gameWithTeams addMove:[turn move] byTeam:teamId];
}

- (void) updateLastPlayerToMove
{
    NSString *username = [[GameStore sharedInstance] username];
    TeamId teamId = [_gameWithTeams teamForUser: username];
    [_gameWithTeams updateLastUserToMove:username onTeam:teamId];
}


- (void) updateNextMoveLabels
{
    if ( [[_gameWithTeams next_turn_team_id] intValue] == WhiteTeam )
        _nextMoveTeamLabel.text = @"White Team";
    else
        _nextMoveTeamLabel.text = @"Black Team";
    
    _nextMovePlayerLabel.text = [_gameWithTeams nextUserToMove];
    
}

- (void) updateBoardInteractionStatus
{
    NSString *username = [[GameStore sharedInstance] username];
    TeamId teamId = [_gameWithTeams teamForUser: username];
    NSLog(@"teamId: %d", teamId);
    NSLog(@"next_turn_team_id: %d", [[_gameWithTeams next_turn_team_id] intValue]);
    
    if ( [[_gameWithTeams next_turn_team_id] intValue] == teamId ) {
        
        NSLog(@"this user's team to move - next user to move:[%@] username:[%@]", [_gameWithTeams nextUserToMove], username);
        
        if ( [[_gameWithTeams nextUserToMove] isEqualToString:username] ) {
            self.boardUIView.userInteractionEnabled = YES;
            NSLog(@"->self.boardUIView.userInteractionEnabled = YES");
        }
        else {
            self.boardUIView.userInteractionEnabled = NO;
            NSLog(@"->self.boardUIView.userInteractionEnabled = NO");
        }
        
    }
    else {
        NSLog(@"not this user's team to move");
        self.boardUIView.userInteractionEnabled = NO;
        NSLog(@"->self.boardUIView.userInteractionEnabled = NO");
    }
    
}


#pragma mark - NSNOTIFICATIONS

- (void) turnCompleted: (NSNotification*)notification
{
    id notificationObject = [notification object];
    if ( [notificationObject isKindOfClass:[Turn class]] ) {

        Turn *turn = (Turn *) notificationObject;
        
        if ( lastMoveMadeByRemotePlayer == NO ) {
            [self saveMoveToDB:turn];
            [self updateLastPlayerToMove];
        }
        lastMoveMadeByRemotePlayer = NO;  // reset this to default value
        
        [self updateBoardInteractionStatus];
        [self updateNextMoveLabels];
        
        NSError* error;
        if (![_gameWithTeams save: &error]) {
            [gAppDelegate showAlert: @"Couldn't save game" error: error fatal: NO];
        }
        
    }
}

#pragma mark - KVO

- (void)observeValueForKeyPath:(NSString *)keyPath
                      ofObject:(id)object
                        change:(NSDictionary *)change
                       context:(void *)context
{
    Game *game = self.boardUIView.game;

    if( object == game ) {
        Player *p = game.winner;
        NSString *msg;
        if( p ) {
            msg = [NSString stringWithFormat:@"%@ wins", p.name];
            
            // TODO: update game board state
            
        } else {
            p = game.currentPlayer;
            
            msg = [NSString stringWithFormat:@"Your turn, %@", p.name];
            NSLog(@"%@",msg);
        }
        
        if( game.winner ) {
            UIAlertView *alert;
            alert = [[UIAlertView alloc] initWithTitle: msg
                                               message: @"Congratulations!"
                                              delegate:self
                                     cancelButtonTitle:@"OK"
                                     otherButtonTitles:nil];
            [alert show];
        }
    }
    else if ( object == _gameWithTeams ) {
        
        // FIXME: compare the moves history of the board to the move history
        // stored in the db, and "fast forward" the board and apply all the moves that need
        // to be applied.
        
        if ( [game respondsToSelector:@selector(applyMoveString:)] ) {
            
            NSString *username = [[GameStore sharedInstance] username];
            TeamId teamId = [_gameWithTeams teamForUser: username];
            
            if ( [[_gameWithTeams next_turn_team_id] intValue] == teamId ) {
                // if its our turn, then the other team must have just made a move,
                // therefore, lets apply it to the board
                
                NSDictionary *lastTurn = [[_gameWithTeams turnsNonNull] lastObject];
                
                if ( lastTurn != nil ) {
                    lastMoveMadeByRemotePlayer = YES;
                    NSString *lastMove = [lastTurn objectForKey:@"move"];
                    BOOL moveApplied = [game applyMoveString:lastMove];
                    if ( moveApplied == NO )
                        NSLog(@"warning: move %@ was not applied to board", lastMove);
                    
                }

            }
            
        }
        
        [self updateBoardInteractionStatus];
        
    }
    
}

#pragma mark -
#pragma mark Class Methonds

+ (NSArray *)fastForwardTurns:(NSArray *)gameTurns databaseTurns:(NSArray *)databaseTurns {

    return [NSArray array];
}


@end
