/*  This code is based on Apple's "GeekGameBoard" sample code, version 1.0.
 http://developer.apple.com/samplecode/GeekGameBoard/
 Copyright © 2013 Traun Leyden. All Rights Reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted
 provided that the following conditions are met:
 
 * Redistributions of source code must retain the above copyright notice, this list of conditions
 and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of
 conditions and the following disclaimer in the documentation and/or other materials provided
 with the distribution.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRI-
 BUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "ChessGame.h"
#import "Grid.h"
#import "Piece.h"
#import "QuartzUtils.h"
#import "GGBUtils.h"

typedef enum {
    TagKing = 0,
    TagQueen,
    TagBishop,
    TagKnight,
    TagRook,
    TagPawn
} ChessPieceTag;

typedef enum {
    CastleMoveNone = 0,
    CastleMoveKingside,
    CastleMoveQueenside
} CastleMoveType;

static const NSInteger kNumPlayersPerGame = 2;
static const NSInteger kPlayerBlack = 1;
static const NSInteger kPlayerWhite = 0;
static const NSInteger kKingsideCastleDstColumn = 6;
static const NSInteger kQueensideCastleDstColumn = 2;
static const NSInteger kKingDefaultColumn = 4;
static const NSInteger kKingWhiteDefaultRow = 0;
static const NSInteger kKingBlackDefaultRow = 7;
static const NSInteger kKingsideCastleRookSrcColumn = 7;
static const NSInteger kKingsideCastleRookDstColumn = 5;
static const NSInteger kQueensideCastleRookSrcColumn = 0;
static const NSInteger kQueensideCastleRookDstColumn = 3;

static NSDictionary *kShortNameToTag, *kTagToLongName;

@implementation ChessGame

#pragma mark -
#pragma mark INITIALIZATION AND DEALLOCATION

+ (void) initialize {
    
    if( self == [ChessGame class] ) {

        kShortNameToTag = @{
        @"r" : [NSNumber numberWithInt: TagRook],
        @"b" : [NSNumber numberWithInt: TagBishop],
        @"k" : [NSNumber numberWithInt: TagKnight],
        @"g" : [NSNumber numberWithInt: TagKing],
        @"q" : [NSNumber numberWithInt: TagQueen],
        @"p" : [NSNumber numberWithInt: TagPawn]
        };
        [kShortNameToTag retain];
        
        kTagToLongName = @{
        [NSNumber numberWithInt: TagRook] : @"rook",
        [NSNumber numberWithInt: TagBishop] : @"bishop",
        [NSNumber numberWithInt: TagKnight] : @"knight",
        [NSNumber numberWithInt: TagKing] : @"king",
        [NSNumber numberWithInt: TagQueen] : @"queen",
        [NSNumber numberWithInt: TagPawn] : @"pawn",
        };
        [kTagToLongName retain];

    }
}

- (id) init
{
    self = [super init];
    if (self != nil)
        [self setNumberOfPlayers: kNumPlayersPerGame];
    return self;
}

#pragma mark -
#pragma mark SUPERCLASS OVERRIDES

- (void) setUpBoard
{
    
    RectGrid *board = [[RectGrid alloc] initWithRows: 8 columns: 8 frame: _table.bounds];
    _board = board;
    [_table addSublayer: _board];
    CGPoint pos = _board.position;
    pos.x = floor((_table.bounds.size.width-board.frame.size.width)/2);
    board.position = pos;
    board.allowsMoves = YES;
    board.allowsCaptures = YES;
    board.cellColor = CreateGray(0.0, 0.5);
    board.altCellColor = CreateGray(1.0, 0.25);
    board.lineColor = nil;
    board.reversed = ! [[self.players objectAtIndex: 0] isLocal];
    
    for( int row=0; row<8; row++ ) {
        for( int col=0; col<8; col++ )
            [_board addCellAtRow: row column: col];
    }
    
    [_board release]; // its superlayer still retains it
}

- (NSString*) initialStateString            {return @"rkbqgbkrpppppppp--------------------------------PPPPPPPPRKBQGBKR";}
- (NSString*) stateString                   {return _board.stateString;}
- (void) setStateString: (NSString*)state   {_board.stateString = state;}

- (Piece*) makePieceNamed: (NSString*)name
{
    Piece *piece;
    if ([name isEqualToString: @"-"])
        return nil;
    else {
        if ([[name lowercaseString] isEqualToString: name])
            piece = [self pieceForPlayer: kPlayerWhite withName: name];  // lowercase -> white
        else
            piece = [self pieceForPlayer: kPlayerBlack withName: name]; // uppercase -> black
    }
    return piece;
}

- (Piece*) pieceForPlayer: (int)playerNum withName: (NSString*)name
{
    Piece *p = [[Piece alloc] init];
    p.bounds = CGRectMake(0,0,floor(_board.spacing.width),floor(_board.spacing.height));
    p.style = [self styleForPlayer: playerNum pieceName: name];
    p.owner = [self.players objectAtIndex: playerNum];
    p.name = name;
    p.tag = [self tagForPieceName:name];
    [self _transformPiece: p];
    return [p autorelease];
}


- (BOOL) canBit: (Bit*)bit moveFrom: (id<BitHolder>)srcHolder to: (id<BitHolder>)dstHolder
{

    int playerIndex = [[bit owner] index];
    Player *player = [bit owner];
    Square *src=(Square*)srcHolder, *dst=(Square*)dstHolder;
    
    switch ( [bit tag] ) {
            
        case TagBishop:
            
            if ( [self isStraightLineMoveTo: dst fromSrc: src] )  
                return NO;
            
            if ( [src lineToCell: dst inclusive: YES] == nil )
                return NO;
            
            if ( [self friendlyPiecesInLineFrom: src toDst: dst] )
                return NO;
            
            break;

        case TagRook:
            
            if ( [self isStraightLineMoveTo: dst fromSrc: src] == NO )
                return NO;
            
            if ( [src lineToCell: dst inclusive: YES] == nil )
                return NO;
            
            if ( [self friendlyPiecesInLineFrom: src toDst: dst] )
                return NO;
            
            if ( [self isPlayer: player castlingFrom: src toDst: dst] == YES )
                return YES;
            
            break;

        case TagQueen:
            
            if ( [src lineToCell: dst inclusive: YES] == nil )  
                return NO;
            
            if ( [self friendlyPiecesInLineFrom: src toDst: dst] )
                return NO;
            
            break;
            
        case TagKing:
                        
            if ( [self numSquaresMovedFrom: src toDst: dst] != 1 )
                return NO;
            
            if ( [self friendlyPiecesInLineFrom: src toDst: dst] )
                return NO;
            
            break;
        
        case TagKnight:
            
            if ( [dst bit] != nil && [ [dst bit] isFriendly ] )
                return NO;
            
            if ( [self isLShapedMoveFrom: src toDst: dst] == NO ) 
                return NO;
            
            break;
            
        case TagPawn:
            
            // TODO: pawn promotion
            
            if ( playerIndex == kPlayerWhite && src.row == 1 && dst.row == 3
                && src.column == dst.column && dst.bit == nil)  // white fwd 2 sq
                return YES;
            else if ( playerIndex == kPlayerBlack && src.row == 6 && dst.row == 4
                     && src.column == dst.column && dst.bit == nil)  // black fwd 2 sq
                return YES;
            else if ( ( dst==src.fl && src.fl.bit.unfriendly ) || ( dst==src.fr && src.fr.bit.unfriendly ) )   // attacking
                return YES;
            else if( dst==src.f && src.f.bit == nil )  // move forward 1, square empty
                return YES;
            else
                return NO;
            
            break;

            
        default:
            return YES;
    }
    
    return YES;
    
}

- (void) bit: (Bit*)bit movedFrom: (id<BitHolder>)srcHolder to: (id<BitHolder>)dstHolder
{

    Square *src=(Square*)srcHolder, *dst=(Square*)dstHolder;
    
    Turn *turn = self.currentTurn;
    if( turn.move.length==0 )
        [turn addToMove: src.name];
    [turn addToMove: @"-"];
    [turn addToMove: dst.name];

    Player *player = bit.owner;
    BOOL didCastle = [self isPlayer: player castlingFrom: (Square*)srcHolder toDst: (Square*)dstHolder];
    if ( didCastle )
        [self performCastleMoveForBit: bit movedFrom: (Square*)srcHolder to: (Square*)dstHolder];
    
    [super bit:bit movedFrom:srcHolder to:dstHolder];
    
}

- (Player*) checkForWinner
{
    BOOL foundOpposingKing = NO;
    for( GridCell *cell in _board.cells )
        if ( [cell bit] != nil &&
             [[cell bit] isUnfriendly] &&
             [[cell bit] tag] == TagKing )
            foundOpposingKing = YES;
    
    return foundOpposingKing ? nil : self.currentPlayer;

}

- (void) _transformPiece: (Piece*)piece
{
    CGFloat scale = 1.0;
    piece.transform = CATransform3DMakeScale(scale, scale/cos(self.tablePerspectiveAngle), scale);
    piece.anchorPoint = CGPointMake(0.5, 0.5*cos(self.tablePerspectiveAngle));
}

- (BOOL) canOpponentMoveFrom: (GridCell*)src
{
    return YES;  
}


- (BOOL) applyMoveString: (NSString*)move
{
    GridCell *src = nil;
    for( NSString *ident in [move componentsSeparatedByString: @"-"] ) {
        while( [ident hasSuffix: @"!"] || [ident hasSuffix: @"*"] )
            ident = [ident substringToIndex: ident.length-1];
        GridCell *dst = [_board cellWithName: ident];
        if( dst == nil )
            return NO;
        if( src && ! [self animateMoveFrom: src to: dst] )
            return NO;
        src = dst;
    }
    return YES;
}


#pragma mark -
#pragma mark PRIVATE METHODS


- (void) performCastleMoveForBit: (Bit*)bit movedFrom: (Square *)src to: (Square *)dst {

    Player *player = [bit owner];
    int newKingSquareRow = [self kingsRowForPlayer: player];
    int newKingSquareColumn = -1;
    
    CastleMoveType castleMoveType = [self castleMoveTypeForPlayer: player
                                                             from: src
                                                            toDst: dst];
    switch (castleMoveType) {
        case CastleMoveKingside:
            newKingSquareColumn = kKingsideCastleDstColumn;
            break;
        case CastleMoveQueenside:
            newKingSquareColumn = kQueensideCastleDstColumn;
            break;
        default:
            break;
    }
    
    GridCell *currentKingSquare = [self defaultKingSquareForPlayer: player];
    GridCell *newKingSquare = [_board cellAtRow: newKingSquareRow column: newKingSquareColumn];
    Bit *king = [currentKingSquare bit];
    [newKingSquare setBit: king];
    [currentKingSquare setBit: nil];
    
}

- (BOOL) isKingInStartingPosition: (Player*)player {

    GridCell *kingSquare = [self defaultKingSquareForPlayer: player];
    return [kingSquare bit] != nil && [[kingSquare bit] tag] == TagKing;
    
}

- (int) kingsRowForPlayer: (Player*)player {
    
    int row = -1;
    
    switch ([player index]) {
        case kPlayerWhite:
            row = kKingWhiteDefaultRow;
            break;
        case kPlayerBlack:
            row = kKingBlackDefaultRow;
            break;
        default:
            break;
    }

    return row;
}

- (GridCell *) defaultKingSquareForPlayer: (Player*)player {
    int row = [self kingsRowForPlayer: player];
    return [_board cellAtRow: row column: kKingDefaultColumn];
}

- (CastleMoveType) castleMoveTypeForPlayer: (Player*)player from: (Square *)src toDst: (Square *)dst {
    
    BOOL isKingsideCastling = NO;
    BOOL isQueensideCastling = NO;
    int row = [self kingsRowForPlayer:player];
    
    if ( [self isKingInStartingPosition:player] == NO )
        return CastleMoveNone;

    isKingsideCastling = (src.row == row &&
                          dst.row == row &&
                          src.column == kKingsideCastleRookSrcColumn &&
                          dst.column == kKingsideCastleRookDstColumn);
    
    isQueensideCastling = (src.row == row &&
                           dst.row == row &&
                           src.column == kQueensideCastleRookSrcColumn &&
                           dst.column == kQueensideCastleRookDstColumn);
    
    if ( isKingsideCastling )
        return CastleMoveKingside;
    else if ( isQueensideCastling )
        return CastleMoveQueenside;
    else
        return CastleMoveNone;
    
}

- (BOOL) isPlayer: (Player*)player castlingFrom: (Square *)src toDst: (Square *)dst {
    
    CastleMoveType castleMoveType = [self castleMoveTypeForPlayer: player
                                                             from: src
                                                            toDst: dst];
    return castleMoveType != CastleMoveNone;
    
}

- (BOOL) isLShapedMoveFrom: (Square *)src toDst: (Square *)dst {
    return ( [self didMoveHorizontalBy: 2 from: src toDst: dst] &&
             [self didMoveVerticalBy: 1 from: src toDst: dst] ) ||
           ( [self didMoveHorizontalBy: 1 from: src toDst: dst] &&
             [self didMoveVerticalBy: 2 from: src toDst: dst] );
    
}

- (BOOL) didMoveHorizontalBy:(int)x from:(Square *)src toDst:(Square *)dst {
    return abs ( src.column - dst.column ) == x;
}

- (BOOL) didMoveVerticalBy:(int)x from:(Square *)src toDst:(Square *)dst { 
    return abs ( src.row - dst.row ) == x;
}

- (int) numSquaresMovedFrom: (Square *)src toDst: (Square *)dst {
    NSArray *line = [src lineToCell: dst inclusive: YES];
    return [line count] - 1;  // subtract one to ignore source Square
}

- (BOOL) isStraightLineMoveTo: (Square *)dst fromSrc: (Square *)src {
    return ( src.row == dst.row || src.column == dst.column );
}

- (BOOL) friendlyPiecesInLineFrom: (Square *)src toDst: (Square *)dst
{
    NSArray *line;
    line = [src lineToCell: dst inclusive: YES];
    for ( Square *square in line ) {
        if ( square == src )
            continue;
        Bit *bit = [square bit];
        if ( bit.friendly )
            return YES;
    }
    return NO;
}

- (NSDictionary*) styleWithImage: (NSString*)imageName {
    
    return @{
        @"contents" : (id)GetCGImageNamed(imageName),
        @"contentsGravity" : kCAGravityResizeAspect,
        @"minificationFilter" : kCAFilterLinear
    };
    
}

- (NSInteger) tagForPieceName: (NSString*)name
{
    name = [name lowercaseString];
    return [[kShortNameToTag objectForKey:name] integerValue];
}

- (NSDictionary*) styleForPlayer: (int)playerNum pieceName: (NSString*)name
{
    NSInteger tag = [self tagForPieceName: name];
    NSString *longName = [kTagToLongName objectForKey:[NSNumber numberWithInt:tag]];
    NSString *colorName;
    
    switch (playerNum) {
        case kPlayerWhite:
            colorName = @"white";
            break;
        case kPlayerBlack:
            colorName = @"black";
            break;
        default:
            return nil;
    }
    
    NSString *imageName = [NSString stringWithFormat:@"%@_%@.png", longName, colorName];
    return [self styleWithImage: imageName];
    
}

@end
